using System;
using System.Threading.Tasks;
using FluentAssertions;
using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.Routing;
using Microsoft.AspNetCore.Mvc.ViewFeatures;
using NSubstitute;
using NSubstitute.ExceptionExtensions;
using Website.Controllers;
using Website.Data;
using Website.Models.Auth;
using Website.ViewModels;
using Xunit;

namespace Website.Tests.Controllers {
	public class AccountControllerTests {
		[Fact]
		public void Login_WithReturnUrl_ReturnsViewThatHasModelWithReturnUrlAndFalseFailedAttempt() {
			var authenticationProvider = Substitute.For<IAuthenticationProvider>();
			var apiKeyManager = Substitute.For<IApiKeyManager>();
			var mailboxesApi = Substitute.For<IMailboxesApi>();
			var controller = new AccountController(authenticationProvider, apiKeyManager, mailboxesApi);

			var expected = new LoginViewModel {
				ReturnUrl = "returnUrl",
				FailedAttempt = false
			};

			var result = controller.Login("returnUrl");
			result.Should().BeOfType<ViewResult>();
			(result as ViewResult)?.Model.Should().BeEquivalentTo(expected);
		}

		[Fact]
		public void Login_WithReturnUrlAndFailedAttempt_ReturnsViewThatHasModelWithReturnUrlAndTrueFailedAttempt() {
			var authenticationProvider = Substitute.For<IAuthenticationProvider>();
			var apiKeyManager = Substitute.For<IApiKeyManager>();
			var mailboxesApi = Substitute.For<IMailboxesApi>();
			var controller = new AccountController(authenticationProvider, apiKeyManager, mailboxesApi);

			var expected = new LoginViewModel {
				ReturnUrl = "returnUrl",
				FailedAttempt = true
			};

			var result = controller.Login("returnUrl", true);
			result.Should().BeOfType<ViewResult>();
			(result as ViewResult)?.Model.Should().BeEquivalentTo(expected);
		}

		[Fact]
		public async Task Login_WithLoginRequest_WhenAuthenticationIsUnsuccessful_ReturnsViewThatHasModelWithReturnUrlAndTrueFailedAttempt() {
			var authenticationProvider = Substitute.For<IAuthenticationProvider>();
			var apiKeyManager = Substitute.For<IApiKeyManager>();
			var mailboxesApi = Substitute.For<IMailboxesApi>();
			var controller = new AccountController(authenticationProvider, apiKeyManager, mailboxesApi);

			var request = new LoginRequest {
				Username = "username",
				Password = "password",
				ReturnUrl = "returnUrl"
			};

			authenticationProvider.Authenticate(request).Returns(null as User);

			var expected = new LoginViewModel {
				ReturnUrl = "returnUrl",
				FailedAttempt = true
			};

			var result = await controller.Login(request);
			result.Should().BeOfType<ViewResult>();
			(result as ViewResult)?.Model.Should().BeEquivalentTo(expected);
		}

		[Fact]
		public async Task Login_WithLoginRequestWithReturnUrl_WhenAuthenticationSuccessful_SignsInAndRedirectsToReturnUrl() {
			var authenticationService = Substitute.For<IAuthenticationService>();
			var serviceProvider = Substitute.For<IServiceProvider>();
			serviceProvider.GetService(typeof(IAuthenticationService)).Returns(authenticationService);

			var authenticationProvider = Substitute.For<IAuthenticationProvider>();
			var apiKeyManager = Substitute.For<IApiKeyManager>();
			var mailboxesApi = Substitute.For<IMailboxesApi>();
			var controller = new AccountController(authenticationProvider, apiKeyManager, mailboxesApi) {
				ControllerContext = new ControllerContext {
					HttpContext = new DefaultHttpContext {
						RequestServices = serviceProvider
					}
				}
			};

			var request = new LoginRequest {
				Username = "username",
				Password = "password",
				ReturnUrl = "returnUrl"
			};

			authenticationProvider.Authenticate(request).Returns(new User{Username = "username", Password = "password"});
			
			var result = await controller.Login(request);
			result.Should().BeOfType<RedirectResult>();
			(result as RedirectResult).Url.Should().Be("returnUrl");
		}

		[Fact]
		public async Task Login_WithLoginRequestWithoutReturnUrl_WhenAuthenticationSuccessful_SignsInAndRedirectsToIndex() {
			var serviceProvider = Substitute.For<IServiceProvider>();

			var authenticationService = Substitute.For<IAuthenticationService>();
			serviceProvider.GetService(typeof(IAuthenticationService)).Returns(authenticationService);

			var tempDataDictionaryFactory = Substitute.For<ITempDataDictionaryFactory>();
			serviceProvider.GetService(typeof(ITempDataDictionaryFactory)).Returns(tempDataDictionaryFactory);

			var urlHelperFactory = Substitute.For<IUrlHelperFactory>();
			serviceProvider.GetService(typeof(IUrlHelperFactory)).Returns(urlHelperFactory);

			var authenticationProvider = Substitute.For<IAuthenticationProvider>();
			var apiKeyManager = Substitute.For<IApiKeyManager>();
			var mailboxesApi = Substitute.For<IMailboxesApi>();
			var controller = new AccountController(authenticationProvider, apiKeyManager, mailboxesApi) {
				ControllerContext = new ControllerContext {
					HttpContext = new DefaultHttpContext {
						RequestServices = serviceProvider
					}
				}
			};

			var request = new LoginRequest {
				Username = "username",
				Password = "password"
			};

			authenticationProvider.Authenticate(request).Returns(new User{Username = "username", Password = "password"});
			
			var result = await controller.Login(request);
			result.Should().BeOfType<RedirectToActionResult>();
			(result as RedirectToActionResult).ActionName.Should().Be("Index");
		}

		[Fact]
		public async Task Login_WithLoginRequest_WhenAuthenticationServiceFails_ReturnsViewThatHasModelWithReturnUrlAndTrueFailedAttempt() {
			var authenticationProvider = Substitute.For<IAuthenticationProvider>();
			var apiKeyManager = Substitute.For<IApiKeyManager>();
			var mailboxesApi = Substitute.For<IMailboxesApi>();
			var controller = new AccountController(authenticationProvider, apiKeyManager, mailboxesApi);

			var request = new LoginRequest {Username = "username", Password = "password", ReturnUrl = "returnUrl"};

			authenticationProvider.Authenticate(request).Throws<Exception>();

			var expected = new LoginViewModel {ReturnUrl = "returnUrl", FailedAttempt = true};

			var result = await controller.Login(request);
			result.Should().BeOfType<ViewResult>();
			(result as ViewResult)?.Model.Should().BeEquivalentTo(expected);
		}
	}
}