﻿using System.Security.Claims;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authentication.Cookies;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using Website.Data;
using Website.Models.Auth;
using Website.Models.Mailboxes;
using Website.ViewModels;

namespace Website.Controllers {
	public class AccountController:Controller {
		private readonly IAuthenticationProvider _authenticationProvider;
		private readonly IApiKeyManager _apiKeyManager;
		private readonly IMailboxesApi _mailboxesApi;

		public AccountController(IAuthenticationProvider authenticationProvider, IApiKeyManager apiKeyManager, IMailboxesApi mailboxesApi) {
			_authenticationProvider = authenticationProvider;
			_apiKeyManager = apiKeyManager;
			_mailboxesApi = mailboxesApi;
		}

		[Authorize]
		public async Task<IActionResult> Index() {
			var keys = await _apiKeyManager.List();
			var aliases = await _mailboxesApi.ListAliases();
			var model = new AccountViewModel(keys, aliases);
			return View(model);
		}

		[HttpGet]
		public IActionResult Login(string returnUrl, bool failedAttempt = false) {
			var model = new LoginViewModel {
				ReturnUrl = returnUrl,
				FailedAttempt = failedAttempt
			};
			return View(model);
		}

		[HttpPost]
		public async Task<IActionResult> Login(LoginRequest request) {
			try {
				var user = await _authenticationProvider.Authenticate(request);
				return user != null
					? await SetIdentityAndRedirect(request.ReturnUrl, user)
					: Login(request.ReturnUrl, true);
			}
			catch {
				return Login(request.ReturnUrl, true);
			}
		}

		private async Task<IActionResult> SetIdentityAndRedirect(string returnUrl, User user) {
			var identity = new ClaimsIdentity(CookieAuthenticationDefaults.AuthenticationScheme);
			identity.AddClaim(new Claim(ClaimTypes.Name, user.Username));
			identity.AddClaim(new Claim(ClaimTypes.Email, user.Username));

			var principal = new ClaimsPrincipal(identity);
			await HttpContext.SignInAsync(CookieAuthenticationDefaults.AuthenticationScheme, principal);

			return string.IsNullOrEmpty(returnUrl)
				? (IActionResult) RedirectToAction(nameof(Index))
				: Redirect(returnUrl);
		}

		public async Task<IActionResult> CreateApiKey(string name) {
			await _apiKeyManager.Create(name);
			return RedirectToAction(nameof(Index));
		}

		public async Task<IActionResult> EnableApiKey(string key) {
			await _apiKeyManager.Enable(key);
			return RedirectToAction(nameof(Index));
		}

		public async Task<IActionResult> DisableApiKey(string key) {
			await _apiKeyManager.Disable(key);
			return RedirectToAction(nameof(Index));
		}

		public async Task<IActionResult> DeleteApiKey(string key) {
			await _apiKeyManager.Delete(key);
			return RedirectToAction(nameof(Index));
		}

		public async Task<IActionResult> EnableAlias(int id) {
			await _mailboxesApi.EnableAlias(id);
			return RedirectToAction(nameof(Index));
		}

		public async Task<IActionResult> DisableAlias(int id) {
			await _mailboxesApi.DisableAlias(id);
			return RedirectToAction(nameof(Index));
		}

		public async Task<IActionResult> DeleteAlias(int id) {
			await _mailboxesApi.DeleteAlias(id);
			return RedirectToAction(nameof(Index));
		}

		public async Task<IActionResult> CreateAlias(string source, string destination, bool active) {
			var alias = new NewAlias {Source = source, Destination = destination, Active = active};
			await _mailboxesApi.CreateAlias(alias);
			return RedirectToAction(nameof(Index));
		}
	}
}